
/*	========================================================================================
   Scriptographer script - Shape Contour
   
   V2 : using intersections rather than hits
   
  | Gareth Williams at West Dale Design   May 2012  : gw@westdale.demon.co.uk
  | This is 'thank you' ware : email me if you find it useful.
  | Free for personal and commercial use. 
  | You may modify the script but keep these credit lines. 
  | If a bug fix is found, please email and I will release a new version. 
   
  Run script with two selected paths.
  The upper item is used as a clipping mask, the lower is scanned to see if a point is inside or outside.
  Works with simple and (reasonably) complex paths/groups - and text converted to paths. 

  Can be run multiple times with the same selection and different parameters for odd effects

  Parameter values shown in dialogue: 
		Height: the change in the output Y value when 'inside' the target shape
		Step: the distance between rows scanned - the resolution
		Stroke: of output path - stroke and fill colour can be applied after running the script
		Filled:  adds a small area under each contour line and fills it.
		Fill depth: depth of the fill (!) - 0 is a default meaning 2 * height
		Slope %: from vertical step to a shallow gradient ( very shallow gradients will behave oddly)
		Smoothing: 1 is default - additional smoothing for outer edges and large gaps
		          - negative smoothing behaves oddly - but not prevented
		Height adjust: reduce height for narrow peaks

================================================================================================= */

// Define default values for palette
var values = {
	height: 12,      
	step: 10,
	stroke: 1,
	slope: 50,
	smooth: 1,
	filled: false,    
	fillDepth: 0,    // filled/0 is default depth - not the same as unfilled !
	yAdjust:true     // reduce y for narrow peaks and dips
};

// Define the palette components to edit each value:
var components = {
	
	height: {
		type: 'number', label: 'Height', units: 'point'
	},
	step: {
		type: 'number', label: 'step', units: 'point'
	},
	slope: {
		type: 'number', label: 'Slope %', units: 'percent',range: [0, 100]
	},
	smooth: {
	    type: 'number', label: 'ends smooth'
	},    
    yAdjust:{
	    type: 'checkbox', label: 'height adjust'
	},
	stroke: {
		type: 'number', label: 'Stroke', units: 'point'
	},
	filled: {
	    type: 'checkbox', label: 'Filled'
	},
	fillDepth: {
	    type: 'number', label: 'Fill Depth', units: 'point'
    },
	// The create button, with an onClick() handler that creates the rectangle
	scan: {
		type: 'button', value: 'Scan',
		onClick: function() {
			// run the scan
			scan();
	        }
	}
};  

// globals for slope to avoid recalc
var slopeToRadians = Math.PI/200;
var deltaX = 0;

// run the script ....
var palette = new Palette('Shape Contours', components, values);


//-- END MAIN ------------------------------------------------------

//---- scan ------------------------------------------------------

function scan () {  

// if two items selected  ok - else exit 
if (document.selectedItems.length != 2 ) {
   print('Need 2 items for processing - target item and clipping shape above it\n');
   return;
}
var item1 = document.selectedItems[0];
var item2 = document.selectedItems[1];

if ( item1.isAbove(item2) ) {
    clipItem = item1;
    hitItem = item2;
}
else {
    clipItem = item2;
    hitItem = item1;
}

//Get bounding rect of top item - which is what is scanned
var top = clipItem.bounds.y;  // zero is at top left
var left = clipItem.bounds.x; 
var bottom = clipItem.bounds.y + clipItem.bounds.height;
var right = clipItem.bounds.x + clipItem.bounds.width;

//grab the parameters for efficiency
var step = components.step.value;
var slope = components.slope.value;
var height = components.height.value; 
var smooth = components.smooth.value;
var stroke = components.stroke.value;

var outY;   // calculated output y of each point
var firstY; // calc output y  at left edge
var prevY1; // record previous output y to calculate smoothing
var prevY2; // record previous output y to calculate smoothing
var hit;    // result of hit test

var hitItemType = hitItem.toString().split(' ')[0];
if (hitItemType == 'Path' || hitItemType == 'CompoundPath' || hitItemType == 'Group') {
}else {
   print('Cant process this type of item', hitItemType);
   print('The target should be a closed path, compound path, group of closed paths or text converted to outlines');
   return;
}

if ( ! checkItemClosed(hitItem, hitItemType ) ) {
   print('Warning there may be open path or complex item within the target');
}

// delta x corresponds to the slope 
if (slope == 0 || slope == 100) {
   deltaX = 0 ; // avoid large values 
}else{
  deltaX = height / Math.tan(slope * slopeToRadians);
}

// Add  new group item to active layer to collect the resulting lines
var linegroup = Group();
document.activeLayer.appendBottom(linegroup);

// Scan loop top to bottom
for (y = top; y < bottom; y += step) {

 	var inside = checkContains(hitItem, hitItemType, new Point(left,y)); // set true and false as we move across the hitItem
	var line = Path.Line(left,y,right,y);	
	
	// check if the line intersects the target item - y used for quick check on bounds of item
    var locations = getIntersections(line, y, hitItem);    
    
    // if all inside or all outside, use the line
	if (!locations.length) {
		if ( inside ) {
			 // line needs to move to y-height   
			 line.segments[0].point.y = y-height;
			 line.segments[1].point.y = y-height;
		} else {
			// leave line where it is 
		}  
		outLine = line;
		
	} else {
	  outLine = processLocations(locations, left, right, y, height, slope, inside, stroke);	
	  line.remove();	
	}
	
	// set some defaults
	outLine.strokeColor = CMYKColor(0, 0, 0, 1);
	outLine.strokeWidth = stroke;
    outLine.fillColor = null;
	
	// smooth the result (if smoothing is selected)
	// do this before adding the extra points for filling 
	   smoothPath(outLine,step,slope,smooth)
	
	// now add (optionally) more points at right/y+ and left/y+ to close the line and fill 
	// defaults to 2*height if zero
	// add 2*stroke to put lines outside clip area even if user changes line to be  inside line etc
    if (components.filled.value) {   
	  var lastY  = outLine.segments[outLine.segments.length -1].point.y; // lastSegment does not work ??
	  var fillDepth = components.fillDepth.value;
	  if (!fillDepth) fillDepth = height*2;

	  outLine.add(Segment(right + 2 * stroke, lastY));
	  outLine.add(Segment(right + 2 * stroke, lastY + fillDepth));
	  outLine.add(Segment(left - 2 * stroke, lastY + fillDepth));
	  outLine.add(Segment(left - 2 * stroke, lastY ));
	  outLine.closePath();
	  outLine.fillColor = CMYKColor(0, 0, 0, 0);
    } 

   // put created line into group
   linegroup.appendTop(outLine); 

}; // loop through y

// clip the result 
var clipGroup = Group();
document.activeLayer.appendTop(clipGroup);

clipGroup.appendTop(linegroup);
clipGroup.appendTop(clipItem);
clipGroup.clipped = true;

}; // -- fn scan 


//------ processLocations ------------------------------------------

function processLocations(locations, left, right, y, height, slope, inside, stroke) { 
// returns processed line - one which does have at least one intersection

var oldInside;
var newInside;
var yAdjust = components.yAdjust.value;

	function compareNumbers(a,b) {
	// used by sort to force numeric comparison
	   return(a-b);
    }

    locations.sort(compareNumbers);
    
	if ( inside ) { 
	  y1 = y-height;
	} else {
	  y1 = y;
	}
	var thisX = left - 2 * stroke
	var seg = Segment(Point(thisX,y1));
	var thisLine = Path([ seg ]);
	// stroke width added to avoid showing within crop area for thick lines
	var prevX = left - 2*stroke; 
	var prevY = y1;
	
	
	for ( var i = 0; i < locations.length; i++) {
	
	    var interX = locations[i];	    
	    var x1 = interX - deltaX; // the slope
	    var x2 = interX + deltaX;
	    
	    oldInside = inside;
	    inside = !inside; 
	    
	    if ( inside ) {
		  y1 = y;
		  y2 = y - height;
		} else {
		  y1 = y - height;
		  y2 = y;
	    }

		if (x1 < prevX) {
		    // overlap -- adjust last point to midpoint of overlap
			var lastseg = thisLine.segments[thisLine.segments.length-1];
			var midX = (x1 + prevX) /2;
			lastseg.point.x = midX;  
			var oldHandleIn = lastseg.handleIn.x // negative value between 0 and deltaX
			oldHandleIn = oldHandleIn + (prevX - midX); // moved point left so reduce handle
			var newHandleIn = (midX - x2)/2;    // negative
			lastseg.handleIn.x = Math.max(oldHandleIn, newHandleIn);	//both neg so is min value!
			lastseg.handleOut.x = -lastseg.handleIn.x ;
			
			if (yAdjust)  { // really means adjust y
				if (  inside) {
					lastseg.point.y = y - height + height * (x2 - midX) / (2 * deltaX)
				}else{
					lastseg.point.y = y - height * (x2 - midX) / (2 * deltaX)
				}
			}

		}else{
			// enough space 
			thisLine.add(Segment(x1, y1, 0, 0, deltaX, 0));
		};
	    
	   	thisLine.add(Segment(x2, y2,-deltaX, 0, 0, 0));
	   	
	   	prevX = x2; 
	   	prevY = y2;
	
	}; // for
	
	if (right < prevX) { // last seg
		thisLine.segments[thisLine.segments.length -1].point.x = right;
	}else {
        thisLine.add(Segment(right, y2, 0, 0, 0, 0));       
	}

  
  return thisLine;

}; //-- fn processLine


//------ smoothPath ------------

function smoothPath(path,step,slope,smooth) {
   // takes a path and applies more smoothing
   // to the left and right -- and to any big gaps
   // smoothing of 0 or negative is not well behaved but allowed for odd effects!
   // it's also odd when height is negative

   if (smooth == 1) return; // this has been done
   
   var i;
   var seg0, seg1, seg2, seg3;
   var x0, y0, x1, y1, x2, y2, x3, y3;
   var midX;
   
   
   var smoothSpacing = smooth * step ; 
   var stHandleRatio = 0.5;
   var stdHandle = smooth * step * stHandleRatio ;
   
   // scan points of the line,  4 at a time 
   for (i = 0; i <= path.segments.length-4; i++) {
       var lastIndex = path.segments.length-4
       seg0 = path.segments[i];
       seg1 = path.segments[i+1];
       seg2 = path.segments[i+2];
       seg3 = path.segments[i+3];
      
       x0 = seg0.point.x;
       y0 = seg0.point.y;
       x1 = seg1.point.x;
       y1 = seg1.point.y;
       x2 = seg2.point.x;
       y2 = seg2.point.y;
       x3 = seg3.point.x;
       y3 = seg3.point.y;
       
       
       if        (i==0 & y1 == y0 & y2 < y0 ) {                   // start of object and there is a bump
       
		  if (x1 - x0 > smoothSpacing) {
				seg1.point.x = x2-smoothSpacing;
				seg1.handleOut.x = stdHandle;

		  } else if (x1 - x0 > step ) {
				seg0.handleOut.x = (x2-x0) * stHandleRatio;			 
				path.remove(i+1);
		  }
		  // else leave alone
		  
       } else if ( i ==lastIndex & y1 == y0 & y2 > y0 & y3 == y2 ) {   // end of object
       
			if (x3 - x2 > smoothSpacing) {
				seg2.point.x = x1 + smoothSpacing;
				seg2.handleIn.x = -stdHandle;
			} else if (x3 - x2 > step ) {
			    seg3.handleIn.x = (x1-x3) * stHandleRatio;			
			    path.remove(i+2);			
			}
			// else leave alone
          
       } else if ( y1 >  y0 & y2 == y1 & y3 == y0 ) {                   // dip
       
       		if( x3 - x0 > 2 * smoothSpacing ) {
       		    // move both points
				seg1.point.x = x0 + smoothSpacing;
				seg1.handleIn.x = -stdHandle;
				seg2.point.x = x3 - smoothSpacing;
       		    seg2.handleOut.x =  stdHandle;
       		} else if (x3 - x0 > 2 * step ) {
       		    // use one midpoint
       			midX = (x0+x3)/2;
				seg2.point.x = midX;
				seg2.handleIn.x = (x0 - midX)  * stHandleRatio;
				seg2.handleOut.x = (x3 - midX) * stHandleRatio;
			    path.remove(i+1);   	    }
       
       } else {
          //do nothing
          
       } // case 

   }; // scan of line

   return;
			  
}; //-- fn smoothPath


//--- getIntersections ----------------------------------------------

function getIntersections(line, y, target){    
    
    // speed up the edges by checking against the bounding rectangle of the item
	if ( y < target.bounds.y || y > target.bounds.y + target.bounds.height ) {
	   return [];
     }
     
     var itemType = target.toString().split(' ')[0];  
  
     if (itemType == 'Group') { 
        //recursively process children of group
        var intersections = new Array();
        for (var i=0; i < target.children.length; i++) {  
          var child = target.children[i];
          var childIntersections = getIntersections(line,y,child);
          intersections = intersections.concat(childIntersections);
        }
        return intersections;
        
    }else if ( itemType == 'Path' || itemType == 'CompoundPath' ) {
           var tempArray1 = line.getIntersections(target); 
           // get the x values
           var tempArray2 = new Array();
           for (var n = 0; n < tempArray1.length; n++) {
               tempArray2.push(tempArray1[n].point.x)  
           }
           return tempArray2;
           
	}else {
	    return [];
    }
    
}; // getIntersections


//--- checkItemClosed -----------------------------------------------

function checkItemClosed(item, itemType) {
    // called recursively for groups and compound items
    
    if (itemType == 'Group' || itemType == 'CompoundPath' ) {
    
       for (var i=0; i < item.children.length ; i++) {
       
          var childType = item.children[i].toString().split(' ')[0];
          if (! checkItemClosed(item.children[i], childType ) ) {
             return false;
          }
       }
       
    }else if ( itemType == 'Path' ) {
         if(!item.closed) {
            return false;
         }
	}else {
	     print ('<'+itemType+'> not recognised');
	    return false;
    }
    
    // no open items found
    return true;
    
}; // checkItemClosed

//--- checkItemContains -------------------

function checkContains(item,itemType,point) {
// need this to process groups as they do not have a contains method

// speed up  edges by checking against the bounding rectangle of the item
if ( ! item.bounds.contains(point)) {
   return false;
}
   
if (itemType == 'Group' || itemType == 'CompoundPath' ) {
      
   for (var i=0; i < item.children.length; i++) {
	  var childType = item.children[i].toString().split(' ')[0];
	  if ( checkContains(item.children[i], childType, point ) ) {
		 return true;
	  }
   }
   return false;
       
}else{
    
    return item.contains(point);
}

};//--checkItemContains
